==> ump-0.8.3.2-src.tar.bz2: ump-0.8.3.2/src/math2.variable.cpp <==

/*
 *  Math2
 *  Copyright (c) 2003-2006 by Mattias Hultgren <mattias_hultgren@tele2.se>
 *
 *  See math2.h
 */


#include "math2.h"
#include <string.h>
#include <stdio.h>

namespace math
{

void Variable::append_to_string( std::string &str, const Format &fmt, bool show_name ) const
       throw(error_obj)
{
	std::string tmp_str;

	try
	{
		if( has_name()  &&  show_name )
		{
			str.append( *get_name() );
			tmp_str = " = ";
			str.append( tmp_str );
		}

		switch( get_type() )
		{
		case VariableType_Matrix:
			get_matrix()->append_to_string( str, fmt, (has_name()  &&  show_name) );
			break;

		case VariableType_String:
			tmp_str = "\"";
			str.append( tmp_str );
			str.append( *get_string() );
			str.append( tmp_str );
			break;

		case VariableType_Complex:
			if( value == 0 )
			{
				tmp_str = "void";
				str.append( tmp_str );
			}
			else
				get_complex()->append_to_string( str, fmt );
			break;

		case VariableType_Picture:
			{
				const Picture *pic = get_picture();
				char tmp[100];
				snprintf( tmp, 100, "%s %ldx%ld", _("Picture"), pic->pic.get_width(),
				          pic->pic.get_height() );
				tmp_str = tmp;
				str.append( tmp_str );
			}
			break;

		case VariableType_Array:
			{
				const Array *array = get_array();

				if( array->get_size() > 0 )
				{
					tmp_str = "( ";
					str.append( tmp_str );
					if( array->get_variable(0)->get_type() == VariableType_Complex )
						array->get_variable(0)->get_complex()->append_to_string( str, fmt );
					else if( array->get_variable(0)->get_type() == VariableType_Boolean )
					{
						tmp_str = (array->get_variable(0)->get_boolean() ? "true" : "false");
						str.append( tmp_str );
					}
					else
					{
						tmp_str = VARIABLE_TYPE_NAMES[array->get_variable(0)->get_type()];
						str.append( tmp_str );
					}
				}
				else
				{
					tmp_str = "( ";
					str.append( tmp_str );
					tmp_str = _("Empty array");
					str.append( tmp_str );
				}

				for( uint32 i=1; i<array->get_size(); i++ )
				{
					tmp_str = ", ";
					str.append( tmp_str );
					if( array->get_variable(i)->get_type() == VariableType_Complex )
						array->get_variable(i)->get_complex()->append_to_string( str, fmt );
					else if( array->get_variable(i)->get_type() == VariableType_Boolean )
					{
						tmp_str = (array->get_variable(i)->get_boolean() ? "true" : "false");
						str.append( tmp_str );
					}
					else
					{
						tmp_str = VARIABLE_TYPE_NAMES[array->get_variable(i)->get_type()];
						str.append( tmp_str );
					}
				}

				tmp_str = " )";
				str.append( tmp_str );
			}
			break;

		case VariableType_Boolean:
			tmp_str = (get_boolean() ? "true" : "false");
			str.append( tmp_str );
			break;
		}
	}
	catch(error_obj error) {  throw error;                 }
	catch(...)             {  Throw_Couldnt_get_memory();  }
}


Variable::Variable()
{
	type = VariableType_Boolean;

	isconstant = false;
	can_change_type = true;

	inuse = false;

	boolean = true;

	change_func = 0;
	delete_func = 0;
}
Variable::Variable(const Variable &src) throw(error_obj)
{
	*this = Variable();
	*this = src;
}

Variable::~Variable()
{
	if( delete_func != 0 )
		(*delete_func)(this);
	change_func = 0;
	isconstant = false;
	clear();
}

// this function only copies the value/Matrix/function/list/string of the Variable
void Variable::operator=(const Variable &src) throw(error_obj)
{
	if(&src == this)
		return;

	switch(src.type)
	{
	case VariableType_Boolean:
		set_boolean(src.boolean);
		break;

	case VariableType_Complex:
		if( src.value == 0 )
			set_void();
		else
			set_complex(*src.value);
		break;

	case VariableType_Matrix:
		set_matrix(src.matr);
		break;

	case VariableType_String:
		if(src.str == 0)
			set_string("");
		else
		{
			if(src.code_line != 0)
				set_code_line( src.str->c_str() );
			else
				set_string( *src.str );
		}
		break;

	case VariableType_Array:
		set_array(src.array);
		break;

	case VariableType_Picture:
		set_picture(src.pic);
		break;

	default: // this should only occur when I've forgotten to add a curtain Variable-type to this function
		Throw_Data_type_is_unsupported_by("Variable::operator=");
	}
}

bool Variable::operator==( const Variable &var ) const
{
	if( type != var.type )
		return false;

	else if( type == VariableType_Complex )
	{
		if( is_void() )
			return var.is_void();
		else if( var.is_void() )
			return false;
		else
			return (*get_complex() == *var.get_complex());
	}
	else if( type == VariableType_Boolean )
		return (get_boolean() == var.get_boolean());

	else if( type == VariableType_Matrix )
		return (*get_matrix() == *var.get_matrix());

	else if( type == VariableType_String )
		return (*get_string() == *var.get_string());
	
	else if( type == VariableType_Picture )
		return (*get_picture() == *var.get_picture());
	
	else // if( type == VariableType_Array )
		return (*get_array() == *var.get_array());
}

void Variable::set_to_constant(void)
{
	isconstant = true;
}

void Variable::set_can_change_type(bool boolean) throw(error_obj)
{
	if(isconstant == true)
		Throw_Tried_to_change_a_constant();

	can_change_type = boolean;
}

void Variable::set_name(const std::string &newname) throw(error_obj)
{
	if(isconstant == true)
		Throw_Tried_to_change_a_constant();

	try{  name = newname;  }
	catch(...) {  Throw_Couldnt_get_memory();  }
}

const Matrix* Variable::get_matrix(void) const throw(error_obj)
{
	if(type != VariableType_Matrix)
		THROW_MACRO_FIX( 44, "Variable isn't a 'Matrix'-type" );

	if(matr == 0)
		THROW_MACRO_FIX( 44,  "Variable has a NULL-object" );

	return matr;
}

Matrix* Variable::get_matrix_rw(void) throw(error_obj)
{
	if(type != VariableType_Matrix)
		THROW_MACRO_FIX( 44, "Variable isn't a 'Matrix'-type" );

	if(matr == 0)
		THROW_MACRO_FIX( 44,  "Variable has a NULL-object" );

	if(isconstant == true)
		Throw_Tried_to_change_a_constant();

	return matr;
}

const Array* Variable::get_array(void) const throw(error_obj)
{
	if(type != VariableType_Array)
		THROW_MACRO_FIX( 44, "Variable isn't a 'Array'-type" );

	if(array == 0)
		THROW_MACRO_FIX( 44, "Variable has a NULL-object" );

	return array;
}

Array* Variable::get_array_rw(void) throw(error_obj)
{
	if(type != VariableType_Array)
		THROW_MACRO_FIX( 44, "Variable isn't a 'Array'-type" );

	if(array == 0)
		THROW_MACRO_FIX( 44, "Variable has a NULL-object" );

	if(isconstant == true)
		Throw_Tried_to_change_a_constant();

	return array;
}

const Picture* Variable::get_picture(void) const throw(error_obj)
{
	if(type != VariableType_Picture)
		THROW_MACRO_FIX( 44, "Variable isn't a 'Picture'-type" );

	if(pic == 0)
		THROW_MACRO_FIX( 44, "Variable has a NULL-object" );

	return pic;
}

Picture* Variable::get_picture_rw(void) throw(error_obj)
{
	if(type != VariableType_Picture)
		THROW_MACRO_FIX( 44, "Variable isn't a 'Picture'-type" );

	if(pic == 0)
		THROW_MACRO_FIX( 44, "Variable has a NULL-object" );

	if(isconstant == true)
		Throw_Tried_to_change_a_constant();

	return pic;
}

void Variable::clear(void) throw(error_obj)
{
	if(isconstant == true)
		Throw_Tried_to_change_a_constant();
	switch(type)
	{
	case VariableType_Boolean:
		boolean = true;
		break;

	case VariableType_Complex:
		delete value;
		value = 0;
		break;

	case VariableType_String:
		delete code_line;
		delete str;
		code_line = 0;
		str = 0;
		break;

	case VariableType_Matrix:
		delete matr;
		matr = 0;
		break;

	case VariableType_Array:
		delete array;
		array = 0;
		break;

	case VariableType_Picture:
		delete pic;
		pic = 0;
		break;

	default: // this should only occur when I've forgotten to add a curtain Variable-type to this function
		Throw_Data_type_is_unsupported_by("Variable::clear");
	}
	if(change_func != 0)
		(*change_func)(this);
}

void Variable::signal_changed(void)
{
	if( change_func != 0 )
		(*change_func)( this );
}

void Variable::set_complex(const Complex &newvalue) throw(error_obj)
{
	if(isconstant == true)
		Throw_Tried_to_change_a_constant();

	if(type != VariableType_Complex)
	{
		if(can_change_type == false)
			THROW_MACRO_FIX( 28,  "Tried to change type of a type-locked variable" );

		void (*tmp)(Variable *ptr);
		tmp = change_func;
		change_func = 0;

		clear();

		type = VariableType_Complex;
		value = 0;

		change_func = tmp;
	}

	if(value == 0)
	{
		try{  value = new Complex;  }
		catch(...) {  Throw_Couldnt_get_memory();  }
	}

	*value = newvalue;

	if(change_func != 0)
		(*change_func)(this);
}

void Variable::set_boolean(bool newbool) throw(error_obj)
{
	if(isconstant == true)
		Throw_Tried_to_change_a_constant();
	if((can_change_type == false) & (type != VariableType_Boolean))
		THROW_MACRO_FIX( 28, "Tried to change type of a type-locked variable" );

	clear();

	type  = VariableType_Boolean;
	boolean = newbool;

	if(change_func != 0)
		(*change_func)(this);
}

void Variable::set_array(const Array *newarray) throw(error_obj)
{
	if(isconstant == true)
		Throw_Tried_to_change_a_constant();
	if((can_change_type == false) & (type != VariableType_Array))
		THROW_MACRO_FIX( 28, "Tried to change type of a type-locked variable" );

	if(type == VariableType_Array)
	{
		if(array == newarray) // the array's are the same
			return;
	}
	clear();

	type = VariableType_Array;
	array = 0;

	if(newarray == 0)
		return;

	try{  array = new Array;  }
	catch(...) {  Throw_Couldnt_get_memory();  }

	try{  *array = *newarray;  }
	catch(...)
	{
		clear();
		throw;
	}

	if(change_func != 0)
		(*change_func)(this);
}

void Variable::set_matrix(const Matrix *newmatrix) throw(error_obj)
{
	if(isconstant == true)
		Throw_Tried_to_change_a_constant();
	if((can_change_type == false) & (type != VariableType_Matrix))
		THROW_MACRO_FIX( 28, "Tried to change type of a type-locked variable" );

	if(type == VariableType_Matrix)
	{
		if(matr == newmatrix) // the Matrix' are the same
			return;
	}
	else
	{
		clear();

		type  = VariableType_Matrix;
		matr = 0;

		if(newmatrix == 0)
			return;

		try{  matr = new Matrix;  }
		catch(...) {  Throw_Couldnt_get_memory();  }
	}

	try{  *matr = *newmatrix;  }
	catch(...)
	{
		clear();
		throw;
	}

	if(change_func != 0)
		(*change_func)(this);
}

void Variable::set_picture(const Picture *newpic) throw(error_obj)
{
	if(isconstant == true)
		Throw_Tried_to_change_a_constant();
	if((can_change_type == false) & (type != VariableType_Picture))
		THROW_MACRO_FIX( 28, "Tried to change type of a type-locked variable" );

	if(type == VariableType_Picture)
	{
		if(pic == newpic) // the Picture's are the same
			return;
	}
	else
	{
		clear();

		type = VariableType_Picture;
		pic = 0;

		if(newpic == 0)
			return;

		try{  pic = new Picture;  }
		catch(...) {  Throw_Couldnt_get_memory();  }
	}

	try{  *pic = *newpic;  }
	catch(...)
	{
		clear();
		throw;
	}

	if(change_func != 0)
		(*change_func)(this);
}

const std::string* Variable::get_string(void) const throw(error_obj)
{
	if(type != VariableType_String)
		THROW_MACRO_FIX( 44, "Variable isn't a 'Function/String'-type" );

	if(str == 0)
		THROW_MACRO_FIX( 45, "Variable has a NULL-object" );

	return str;
}

std::string* Variable::get_string_rw(void) throw(error_obj)
{
	if(type != VariableType_String)
		THROW_MACRO_FIX( 44, "Variable isn't a 'Function/String'-type" );

	if(str == 0)
		THROW_MACRO_FIX( 45, "Variable has a NULL-object" );

	if(isconstant == true)
		Throw_Tried_to_change_a_constant();

	return str;
}

const std::string* Variable::get_name(void) const throw(error_obj)
{
	if(name.empty())
		THROW_MACRO_FIX( 45, "Variable has no name" );

	return &name;
}

void Variable::calc_code_line( Variable *res ) throw(error_obj)
{
	bool tmp_isconstant;

	if(type != VariableType_String)
		THROW_MACRO_FIX( 44, "Variable isn't a 'String'-type" );

	if(code_line == 0)
	{
		if(str == 0)
			THROW_MACRO_FIX( 45, "Variable has a NULL-object" );

		try{  code_line = new CodeLine;  }
		catch(...) {  Throw_Couldnt_get_memory();  }

		code_line->set_code_line( str->c_str() );
	}

	if(inuse)
	{
		if(name.empty())
			Throw_Circle_reference_of_variable( " " );
		else
			Throw_Circle_reference_of_variable( name.c_str() );
	}

	if(has_name()) // a nameless Variable can't have a circle reference
		inuse = true;
	tmp_isconstant = isconstant;
	isconstant = true;

	try{  code_line->calc( res );  }
	catch(...)
	{
		inuse = false;
		isconstant = tmp_isconstant;
		throw;
	}
	inuse = false;
	isconstant = tmp_isconstant;
}

void Variable::set_string(const std::string &newstring) throw(error_obj)
{
	if(isconstant == true)
		Throw_Tried_to_change_a_constant();
	if((can_change_type == false) & (type != VariableType_String))
		THROW_MACRO_FIX( 28, "Tried to change type of a type-locked variable" );

	clear();

	type = VariableType_String;
	str = 0;
	code_line = 0;

	try{  str = new std::string(newstring);  }
	catch(...) {  Throw_Couldnt_get_memory();  }

	if(change_func != 0)
		(*change_func)(this);
}

void Variable::set_code_line(const char *new_code_line) throw(error_obj)
{
	void (*tmp_change_func)(Variable *ptr);

	tmp_change_func = change_func;
	change_func = 0;

	set_string( new_code_line );

	change_func = tmp_change_func;

	code_line = 0;

	if(new_code_line == 0)
		return;
	if(new_code_line[0] == 0)
		return;

	try{  code_line = new CodeLine;  }
	catch(...) {  Throw_Couldnt_get_memory();  }


	code_line->set_code_line(new_code_line);


	if(change_func != 0)
		(*change_func)(this);
}

const Complex* Variable::get_complex(void) const throw(error_obj)
{
	if(type != VariableType_Complex)
		THROW_MACRO_FIX( 44, "Variable isn't a 'Complex'-type" );

	if(value == 0)
		THROW_MACRO_FIX( 44, "Variable isn't a 'Complex'-type" ); // this is a void variable

	return value;
}

Complex* Variable::get_complex_rw(void) throw(error_obj)
{
	if(type != VariableType_Complex)
		THROW_MACRO_FIX( 44, "Variable isn't a 'Complex'-type" );

	if(value == 0)
		THROW_MACRO_FIX( 44, "Variable isn't a 'Complex'-type" ); // this is a void variable

	return value;
}

bool Variable::get_boolean(void) const throw(error_obj)
{
	if(type == VariableType_Boolean)
		return boolean;

	THROW_MACRO_FIX( 44, "Variable isn't a 'Boolean'-type" );
}

void Variable::set_void(void)
{
	if(isconstant == true)
		Throw_Tried_to_change_a_constant();

	if(can_change_type == false)
		THROW_MACRO_FIX( 28,  "Tried to change type of a type-locked variable" );

	void (*tmp)(Variable *ptr);
	tmp = change_func;
	change_func = 0;

	clear();

	type = VariableType_Complex;
	value = 0;

	change_func = tmp;

	if(change_func != 0)
		(*change_func)(this);
}

bool Variable::is_void(void) const
{
	if(type != VariableType_Complex)
		return false;

	return (value == 0);
}

} // namespace math

==> ump-0.8.4-src.tar.bz2: ump-0.8.4/src/math2.variable.cpp <==

/*
 *  Math2
 *  Copyright (c) 2003-2006 by Mattias Hultgren <mattias_hultgren@tele2.se>
 *
 *  See math2.h
 */


#include "math2.h"
#include <string.h>
#include <stdio.h>
#include <new>

namespace math
{

void Variable::append_to_string( std::string &str, const Format &fmt, bool show_name ) const
       throw(error_obj)
{
	std::string tmp_str;

	try
	{
		if( has_name()  &&  show_name )
		{
			str.append( *get_name() );
			tmp_str = " = ";
			str.append( tmp_str );
		}

		switch( get_type() )
		{
		case VariableType_Matrix:
			get_matrix()->append_to_string( str, fmt, (has_name()  &&  show_name) );
			break;

		case VariableType_String:
			tmp_str = "\"";
			str.append( tmp_str );
			str.append( *get_string() );
			str.append( tmp_str );
			break;

		case VariableType_Complex:
			get_complex()->append_to_string( str, fmt );
			break;

		case VariableType_Picture:
			{
				const Picture *pic = get_picture();
				char tmp[100];
				snprintf( tmp, 100, "%s %ldx%ld", _("Picture"), pic->pic.get_width(),
				          pic->pic.get_height() );
				tmp_str = tmp;
				str.append( tmp_str );
			}
			break;

		case VariableType_Array:
			{
				const Array *array = get_array();

				if( array->get_size() > 0 )
				{
					tmp_str = "( ";
					str.append( tmp_str );
					if( array->get_variable(0)->get_type() == VariableType_Complex )
						array->get_variable(0)->get_complex()->append_to_string( str, fmt );
					else if( array->get_variable(0)->get_type() == VariableType_Boolean )
					{
						tmp_str = (array->get_variable(0)->get_boolean() ? "true" : "false");
						str.append( tmp_str );
					}
					else
					{
						tmp_str = VARIABLE_TYPE_NAMES[array->get_variable(0)->get_type()];
						str.append( tmp_str );
					}
				}
				else
				{
					tmp_str = "( ";
					str.append( tmp_str );
					tmp_str = _("Empty array");
					str.append( tmp_str );
				}

				for( uint32 i=1; i<array->get_size(); i++ )
				{
					tmp_str = ", ";
					str.append( tmp_str );
					if( array->get_variable(i)->get_type() == VariableType_Complex )
						array->get_variable(i)->get_complex()->append_to_string( str, fmt );
					else if( array->get_variable(i)->get_type() == VariableType_Boolean )
					{
						tmp_str = (array->get_variable(i)->get_boolean() ? "true" : "false");
						str.append( tmp_str );
					}
					else
					{
						tmp_str = VARIABLE_TYPE_NAMES[array->get_variable(i)->get_type()];
						str.append( tmp_str );
					}
				}

				tmp_str = " )";
				str.append( tmp_str );
			}
			break;

		case VariableType_Boolean:
			tmp_str = (get_boolean() ? "true" : "false");
			str.append( tmp_str );
			break;

		case VariableType_Void:
			str.append( "void" );
			break;
		}
	}
	catch(error_obj error) {  throw error;                 }
	catch(...)             {  Throw_Couldnt_get_memory();  }
}


Variable::Variable()
{
	type = VariableType_Boolean;

	isconstant = false;
	can_change_type = true;

	inuse = false;

	boolean = true;

	change_func = 0;
	delete_func = 0;
}
Variable::Variable(const Variable &src) throw(error_obj)
{
	type = VariableType_Boolean;

	isconstant = false;
	can_change_type = true;

	inuse = false;

	boolean = true;

	change_func = 0;
	delete_func = 0;

	*this = src;
}

Variable::~Variable()
{
	if( delete_func != 0 )
		(*delete_func)(this);
	change_func = 0;
	isconstant = false;
	clear();
}

// this function only copies the value/Matrix/function/list/string of the Variable
void Variable::operator=(const Variable &src) throw(error_obj)
{
	if(&src == this)
		return;

	switch(src.type)
	{
	case VariableType_Boolean:
		set_boolean(src.boolean);
		break;

	case VariableType_Complex:
		set_complex( *Complex_ptr(src.complex_buf) );
		break;

	case VariableType_Matrix:
		set_matrix(src.matr);
		break;

	case VariableType_String:
		if(src.str == 0)
			set_string("");
		else
		{
			if(src.code_line != 0)
				set_code_line( src.str->c_str() );
			else
				set_string( *src.str );
		}
		break;

	case VariableType_Array:
		set_array(src.array);
		break;

	case VariableType_Picture:
		set_picture(src.pic);
		break;

	case VariableType_Void:
		set_void();
		break;

	default: // this should only occur when I've forgotten to add a curtain Variable-type to this function
		Throw_Data_type_is_unsupported_by("Variable::operator=");
	}
}

bool Variable::operator==( const Variable &var ) const
{
	if( type != var.type )
		return false;

	else if( type == VariableType_Complex )
		return (*get_complex() == *var.get_complex());

	else if( type == VariableType_Boolean )
		return (get_boolean() == var.get_boolean());

	else if( type == VariableType_Matrix )
		return (*get_matrix() == *var.get_matrix());

	else if( type == VariableType_String )
		return (*get_string() == *var.get_string());
	
	else if( type == VariableType_Picture )
		return (*get_picture() == *var.get_picture());
	
	else  if( type == VariableType_Array )
		return (*get_array() == *var.get_array());

	else // if( type == VariableType_Void )
		return var.is_void();
}

void Variable::set_to_constant(void)
{
	isconstant = true;
}

void Variable::set_can_change_type(bool boolean) throw(error_obj)
{
	if(isconstant == true)
		Throw_Tried_to_change_a_constant();

	can_change_type = boolean;
}

void Variable::set_name(const std::string &newname) throw(error_obj)
{
	if(isconstant == true)
		Throw_Tried_to_change_a_constant();

	try{  name = newname;  }
	catch(...) {  Throw_Couldnt_get_memory();  }
}

const Matrix* Variable::get_matrix(void) const throw(error_obj)
{
	if(type != VariableType_Matrix)
		THROW_MACRO_FIX( 44, "Variable isn't a 'Matrix'-type" );

	if(matr == 0)
		THROW_MACRO_FIX( 44,  "Variable has a NULL-object" );

	return matr;
}

Matrix* Variable::get_matrix_rw(void) throw(error_obj)
{
	if(type != VariableType_Matrix)
		THROW_MACRO_FIX( 44, "Variable isn't a 'Matrix'-type" );

	if(matr == 0)
		THROW_MACRO_FIX( 44,  "Variable has a NULL-object" );

	if(isconstant == true)
		Throw_Tried_to_change_a_constant();

	return matr;
}

const Array* Variable::get_array(void) const throw(error_obj)
{
	if(type != VariableType_Array)
		THROW_MACRO_FIX( 44, "Variable isn't a 'Array'-type" );

	if(array == 0)
		THROW_MACRO_FIX( 44, "Variable has a NULL-object" );

	return array;
}

Array* Variable::get_array_rw(void) throw(error_obj)
{
	if(type != VariableType_Array)
		THROW_MACRO_FIX( 44, "Variable isn't a 'Array'-type" );

	if(array == 0)
		THROW_MACRO_FIX( 44, "Variable has a NULL-object" );

	if(isconstant == true)
		Throw_Tried_to_change_a_constant();

	return array;
}

const Picture* Variable::get_picture(void) const throw(error_obj)
{
	if(type != VariableType_Picture)
		THROW_MACRO_FIX( 44, "Variable isn't a 'Picture'-type" );

	if(pic == 0)
		THROW_MACRO_FIX( 44, "Variable has a NULL-object" );

	return pic;
}

Picture* Variable::get_picture_rw(void) throw(error_obj)
{
	if(type != VariableType_Picture)
		THROW_MACRO_FIX( 44, "Variable isn't a 'Picture'-type" );

	if(pic == 0)
		THROW_MACRO_FIX( 44, "Variable has a NULL-object" );

	if(isconstant == true)
		Throw_Tried_to_change_a_constant();

	return pic;
}

void Variable::clear(void) throw(error_obj)
{
	if(isconstant == true)
		Throw_Tried_to_change_a_constant();
	switch(type)
	{
	case VariableType_Boolean:
		boolean = true;
		break;

	case VariableType_Complex:
		if( boolean )
			Complex_ptr( complex_buf )->~Complex();
		boolean = false;
		break;

	case VariableType_String:
		delete code_line;
		delete str;
		code_line = 0;
		str = 0;
		break;

	case VariableType_Matrix:
		delete matr;
		matr = 0;
		break;

	case VariableType_Array:
		delete array;
		array = 0;
		break;

	case VariableType_Picture:
		delete pic;
		pic = 0;
		break;

	case VariableType_Void:
		break;

	default: // this should only occur when I've forgotten to add a curtain Variable-type to this function
		Throw_Data_type_is_unsupported_by("Variable::clear");
	}
	if(change_func != 0)
		(*change_func)(this);
}

void Variable::signal_changed(void)
{
	if( change_func != 0 )
		(*change_func)( this );
}

void Variable::set_complex(const Complex &newvalue) throw(error_obj)
{
	if(isconstant == true)
		Throw_Tried_to_change_a_constant();

	if(type != VariableType_Complex)
	{
		if(can_change_type == false)
			THROW_MACRO_FIX( 28,  "Tried to change type of a type-locked variable" );

		void (*tmp)(Variable *ptr);
		tmp = change_func;
		change_func = 0;

		clear();

		type = VariableType_Complex;
		boolean = false;

		change_func = tmp;
	}

	if( !boolean )
	{
		try{  new (complex_buf) Complex;  boolean = true;  }
		catch(...) {  Throw_Couldnt_get_memory();  }
	}

	*Complex_ptr(complex_buf) = newvalue;

	if(change_func != 0)
		(*change_func)(this);
}

void Variable::set_boolean(bool newbool) throw(error_obj)
{
	if(isconstant == true)
		Throw_Tried_to_change_a_constant();
	if( can_change_type == false  &&  type != VariableType_Boolean )
		THROW_MACRO_FIX( 28, "Tried to change type of a type-locked variable" );

	clear();

	type  = VariableType_Boolean;
	boolean = newbool;

	if(change_func != 0)
		(*change_func)(this);
}

void Variable::set_array(const Array *newarray) throw(error_obj)
{
	if(isconstant == true)
		Throw_Tried_to_change_a_constant();
	if( can_change_type == false  &&  type != VariableType_Array )
		THROW_MACRO_FIX( 28, "Tried to change type of a type-locked variable" );

	if(type == VariableType_Array)
	{
		if(array == newarray) // the array's are the same
			return;
	}
	clear();

	type = VariableType_Array;
	array = 0;

	if(newarray == 0)
		return;

	try{  array = new Array;  }
	catch(...) {  Throw_Couldnt_get_memory();  }

	try{  *array = *newarray;  }
	catch(...)
	{
		clear();
		throw;
	}

	if(change_func != 0)
		(*change_func)(this);
}

void Variable::set_matrix(const Matrix *newmatrix) throw(error_obj)
{
	if(isconstant == true)
		Throw_Tried_to_change_a_constant();
	if( can_change_type == false  &&  type != VariableType_Matrix )
		THROW_MACRO_FIX( 28, "Tried to change type of a type-locked variable" );

	if(type == VariableType_Matrix)
	{
		if(matr == newmatrix) // the Matrix' are the same
			return;
	}
	else
	{
		clear();

		type  = VariableType_Matrix;
		matr = 0;

		if(newmatrix == 0)
			return;

		try{  matr = new Matrix;  }
		catch(...) {  Throw_Couldnt_get_memory();  }
	}

	try{  *matr = *newmatrix;  }
	catch(...)
	{
		clear();
		throw;
	}

	if(change_func != 0)
		(*change_func)(this);
}

void Variable::set_picture(const Picture *newpic) throw(error_obj)
{
	if(isconstant == true)
		Throw_Tried_to_change_a_constant();
	if( can_change_type == false  &&  type != VariableType_Picture )
		THROW_MACRO_FIX( 28, "Tried to change type of a type-locked variable" );

	if(type == VariableType_Picture)
	{
		if(pic == newpic) // the Picture's are the same
			return;
	}
	else
	{
		clear();

		type = VariableType_Picture;
		pic = 0;

		if(newpic == 0)
			return;

		try{  pic = new Picture;  }
		catch(...) {  Throw_Couldnt_get_memory();  }
	}

	try{  *pic = *newpic;  }
	catch(...)
	{
		clear();
		throw;
	}

	if(change_func != 0)
		(*change_func)(this);
}

const std::string* Variable::get_string(void) const throw(error_obj)
{
	if(type != VariableType_String)
		THROW_MACRO_FIX( 44, "Variable isn't a 'Function/String'-type" );

	if(str == 0)
		THROW_MACRO_FIX( 45, "Variable has a NULL-object" );

	return str;
}

std::string* Variable::get_string_rw(void) throw(error_obj)
{
	if(type != VariableType_String)
		THROW_MACRO_FIX( 44, "Variable isn't a 'Function/String'-type" );

	if(str == 0)
		THROW_MACRO_FIX( 45, "Variable has a NULL-object" );

	if(isconstant == true)
		Throw_Tried_to_change_a_constant();

	return str;
}

const std::string* Variable::get_name(void) const throw(error_obj)
{
	if(name.empty())
		THROW_MACRO_FIX( 45, "Variable has no name" );

	return &name;
}

void Variable::calc_code_line( Variable *res ) throw(error_obj)
{
	bool tmp_isconstant;

	if(type != VariableType_String)
		THROW_MACRO_FIX( 44, "Variable isn't a 'String'-type" );

	if(code_line == 0)
	{
		if(str == 0)
			THROW_MACRO_FIX( 45, "Variable has a NULL-object" );

		try{  code_line = new CodeLine;  }
		catch(...) {  Throw_Couldnt_get_memory();  }

		code_line->set_code_line( str->c_str() );
	}

	if(inuse)
	{
		if(name.empty())
			Throw_Circle_reference_of_variable( " " );
		else
			Throw_Circle_reference_of_variable( name.c_str() );
	}

	if(has_name()) // a nameless Variable can't have a circle reference
		inuse = true;
	tmp_isconstant = isconstant;
	isconstant = true;

	try{  code_line->calc( res );  }
	catch(...)
	{
		inuse = false;
		isconstant = tmp_isconstant;
		throw;
	}
	inuse = false;
	isconstant = tmp_isconstant;
}

void Variable::set_string(const std::string &newstring) throw(error_obj)
{
	if( isconstant == true )
		Throw_Tried_to_change_a_constant();
	if( can_change_type == false  &&  type != VariableType_String )
		THROW_MACRO_FIX( 28, "Tried to change type of a type-locked variable" );

	clear();

	type = VariableType_String;
	str = 0;
	code_line = 0;

	try{  str = new std::string(newstring);  }
	catch(...) {  Throw_Couldnt_get_memory();  }

	if(change_func != 0)
		(*change_func)(this);
}

void Variable::set_code_line(const char *new_code_line) throw(error_obj)
{
	void (*tmp_change_func)(Variable *ptr);

	tmp_change_func = change_func;
	change_func = 0;

	set_string( new_code_line );

	change_func = tmp_change_func;

	code_line = 0;

	if(new_code_line == 0)
		return;
	if(new_code_line[0] == 0)
		return;

	try{  code_line = new CodeLine;  }
	catch(...) {  Throw_Couldnt_get_memory();  }


	code_line->set_code_line(new_code_line);


	if(change_func != 0)
		(*change_func)(this);
}

const Complex* Variable::get_complex(void) const throw(error_obj)
{
	if(type != VariableType_Complex)
		THROW_MACRO_FIX( 44, "Variable isn't a 'Complex'-type" );

	if( !boolean )
		THROW_MACRO_FIX( 45, "Variable has a NULL-object" );

	return Complex_ptr(complex_buf);
}

Complex* Variable::get_complex_rw(void) throw(error_obj)
{
	if(type != VariableType_Complex)
		THROW_MACRO_FIX( 44, "Variable isn't a 'Complex'-type" );

	if( !boolean )
		THROW_MACRO_FIX( 45, "Variable has a NULL-object" );

	return Complex_ptr(complex_buf);
}

bool Variable::get_boolean(void) const throw(error_obj)
{
	if(type == VariableType_Boolean)
		return boolean;

	THROW_MACRO_FIX( 44, "Variable isn't a 'Boolean'-type" );
}

void Variable::set_void(void)
{
	if(isconstant == true)
		Throw_Tried_to_change_a_constant();

	if( can_change_type == false  &&  type != VariableType_Void )
		THROW_MACRO_FIX( 28,  "Tried to change type of a type-locked variable" );

	void (*tmp)(Variable *ptr);
	tmp = change_func;
	change_func = 0;

	clear();

	type = VariableType_Void;

	change_func = tmp;

	if(change_func != 0)
		(*change_func)(this);
}

} // namespace math
